# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging

from aenum import Enum

logger = logging.getLogger(__name__)

__all__ = ['Circle', 'Polygon', 'Energy', 'Time', 'Polarization',
           'PolarizationState']


class Circle(object):

    AXIS_TYPE = 'POSITION'
    NAME = 'CIRCLE'

    def __init__(self, ra, dec, radius):
        """
        Circle must have 3 parameters, RA, Dec, and radius.

        :param ra: float    Right Ascension
        :param dec: float   Declination
        :param radius: float    Radius of the circle
        """

        self.ra = ra
        self.dec = dec
        self.radius = radius

    @staticmethod
    def parse_circle(coordinates):
        """
        Parse a string of RA, Dec, and radius into a Circle.

        :param coordinates: str Space delimited RA, Dec, and radius
        :return: Circle instance
        """

        if not coordinates:
            raise ValueError('Empty or None coordinate value')
        values = coordinates.split()

        # Circle must have 3 values
        if len(values) < 3:
            raise ValueError('Circle requires a RA, Dec, and radius')

        return Circle(float(values[0]), float(values[1]), float(values[2]))


class Polygon(object):

    AXIS_TYPE = 'POSITION'
    NAME = 'POLYGON'

    def __init__(self, vertices):
        """
        A Polygon is at a minimum 3 pairs of coordinates (vertices).

        :param vertices: List   List of float of coordinate pairs
        """

        # Polygon must have a minimum of 6 values (3 pairs of coordinates)
        if not vertices or len(vertices) < 3:
            raise ValueError(
                'Polygon requires a minimum of 3 vertices (3 coordinate pairs)')

        self.vertices = vertices

    @staticmethod
    def parse_polygon(coordinates):
        """
        Parse a string of coordinates into a Polygon.

        :param coordinates: str   Space delimited string of coordinates
        :return: Polygon instance
        """

        if not coordinates:
            raise ValueError('Empty or None coordinate value')

        # Polygon must have a minimum of 6 values (3 pairs of coordinates)
        values = coordinates.split()
        if len(values) < 6:
            raise ValueError('Polygon requires a minimum of 6 coordinates')

        # parse coordinates into pairs
        vertices = []
        for i in range(0, len(values), 2):
            vertices.append([float(values[i]), float(values[i + 1])])
        return Polygon(vertices)


class Energy(object):

    AXIS_TYPE = 'SPECTRAL'
    NAME = 'BAND'

    def __init__(self, lower, upper):
        """
        Create an Energy instance from an interval
        with the given lower and upper bounds.

        :param lower: float Lower energy value
        :param upper: float Upper energy value
        """

        self.lower = lower
        self.upper = upper

    @staticmethod
    def parse_energy(bounds):
        """
        Parse a string of lower and upper bounds into an Energy.

        :param bounds: str   Space delimited string of lower and upper bounds
        :return: Energy instance
        """

        if not bounds:
            raise ValueError('Empty or None bounds value')

        values = bounds.split()
        if len(values) != 2:
            raise ValueError('Energy requires a lower and upper bounds')

        return Energy(float(values[0]), float(values[1]))


class Time(object):

    AXIS_TYPE = 'TEMPORAL'
    NAME = 'TIME'

    def __init__(self, lower, upper):
        """
        Create a Time instance from an interval
        with the given lower and upper bounds.

        :param lower: float Lower time value
        :param upper: float Upper time value
        """

        self.lower = lower
        self.upper = upper

    @staticmethod
    def parse_time(bounds):
        """
        Parse a string of lower and upper bounds into a Time.

        :param bounds: str   Space delimited string of lower and upper bounds
        :return: Time instance
        """

        if not bounds:
            raise ValueError('Empty or None bounds value')

        values = bounds.split()
        if len(values) != 2:
            raise ValueError('Time requires a lower and upper bounds')

        return Time(float(values[0]), float(values[1]))


class Polarization(object):

    AXIS_TYPE = 'POLARIZATION'
    NAME = 'POL'

    def __init__(self, states):
        """
        Create a Polarization instance.

        :param states: List List of PolarizationState
        """

        # Polarization must have at least one state
        if not states:
            raise ValueError('Polarization must have one or more states')

        # validate states
        for state in states:
            if not isinstance(state, PolarizationState):
                raise ValueError('Invalid PolarizationState {}'.format(state))

        self.states = states

    @staticmethod
    def parse_polarization(polarization_states):
        """
        Parse a string of polarization states into a Polarization.

        :param polarization_states: str   Space delimited string
                                          of polarization states
        :return: Polarization instance
        """

        if not polarization_states:
            raise ValueError('Empty or None polarization_states value')

        # validate states
        states = polarization_states.split()
        known_states = []
        unknown_states = []
        valid_states = [e.name for e in PolarizationState]
        for state in states:
            if state not in valid_states:
                unknown_states.append(state)
            else:
                known_states.append(PolarizationState[state])
        if unknown_states:
            raise ValueError('Unknown polarization states {}'
                             .format(unknown_states))

        # Polarization must have at least one state
        if not known_states:
            raise ValueError('Polarization must have one or more states')

        return Polarization(known_states)


class PolarizationState(Enum):
    """
    Enum of polarization states.
    """
    I = 1   # noqa: E741
    Q = 2
    U = 3
    V = 4
    POLI = 5
    FPOLI = 6
    POLA = 7
    EPOLI = 8
    CPOLI = 9
    NPOLI = 10
    RR = -1
    LL = -2
    RL = -3
    LR = -4
    XX = -5
    YY = -6
    XY = -7
    YX = -8



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from cadccutout.transform import Transform
from cadccutout.pixel_cutout_hdu import PixelCutoutHDU
from cadccutout.no_content_error import NoContentError
from cadccutout.file_helpers.base_file_helper import BaseFileHelper
from astropy.wcs import WCS
from astropy.io import fits
from astropy.io.fits import PrimaryHDU


import logging


__all__ = ['FITSHelper']

logger = logging.getLogger(__name__)


class FITSHelper(BaseFileHelper):

    def __init__(self, input_stream, output_writer):
        """
        Create a new BaseFileHelper used for different file types.  Concrete
         instances are expected to extend this
        class to provide some common state.

        :param input_stream:    The Reader to read the file data from.
        :param output_writer:   The Writer to write the cutout to.
        """
        super(FITSHelper, self).__init__(input_stream, output_writer)

    def _post_sanitize_header(self, header, cutout_result):
        """
        Fix the CRPIX offset
        """
        if cutout_result.wcs is not None:
            wcs = cutout_result.wcs.wcs
            if wcs.crpix is not None:
                cutout_crpix = wcs.crpix

                for idx, val in enumerate(cutout_crpix):
                    header_key = 'CRPIX{}'.format(idx + 1)
                    curr_val = header.get(header_key)
                    logger.debug(
                        'Adjusting {} from {} to {}'.format(
                            header_key, curr_val, val))
                    header.set('CRPIX{}'.format(idx + 1), val)

            if wcs.has_pc():
                pc = wcs.pc
                for i in range(wcs.naxis):
                    for j in range(wcs.naxis):
                        header.set('PC{}_{}'.format(i + 1, j + 1), pc[i][j])
            elif wcs.has_cd():
                cd = wcs.cd
                for i in range(wcs.naxis):
                    for j in range(wcs.naxis):
                        header.set('CD{}_{}'.format(i + 1, j + 1), cd[i][j])

    def _get_wcs(self, header):
        naxis_value = header.get('NAXIS')

        if naxis_value is not None and int(naxis_value) > 0:
            for nax in range(1, naxis_value + 1):
                ctype = header.get('CTYPE{0} '.format(nax))
                if ctype is not None and ctype.endswith('-SIP'):
                    naxis = 2
                    break
                else:
                    naxis = None
        else:
            naxis = naxis_value

        return WCS(header=header, naxis=naxis, fix=False)

    def _pixel_cutout(self, hdu, cutout_dimension):
        extension = cutout_dimension.get_extension()
        header = hdu.header
        wcs = self._get_wcs(header)

        logger.debug('Cutting out from extension {}'.format(extension))

        cutout_result = self.do_cutout(
            data=hdu.data, cutout_dimension=cutout_dimension, wcs=wcs)

        self._post_sanitize_header(header, cutout_result)
        result = fits.ImageHDU(header=header, data=cutout_result.data)
        self._fix_header(result.header, header)
        return result

    def _fix_header(self, target, original):
        # overcome astropy's tendency to remove keywords from original headers
        if 'BSCALE' in original:
            target['BSCALE'] = original['BSCALE']
        if 'BZERO' in original:
            target['BZERO'] = original['BZERO']

    def _add_primary_hdu(self, hdu_list, result_hdu_list):
        if isinstance(hdu_list[0], PrimaryHDU):
            logger.debug('Appending Primary HDU.')
            return fits.HDUList([hdu_list[0]])
        else:
            logger.debug('HDU List does NOT contain a Primary HDU.'
                         'Skipping.')
            return result_hdu_list

    def _require_primary_hdu(self, cutout_dimensions):
        # returns True if resulting cutout requires primary HDU from the
        # original hdu list. This is the case when cutouts are done in
        # different extensions of the file
        last_ext = -1
        for c in cutout_dimensions:
            if last_ext != -1 and last_ext != c._extension:
                return True
            last_ext = c._extension
        return False

    def _check_hdu_list(self, cutout_dimensions, hdu_list):
        has_match = False
        len_cutout_dimensions = len(cutout_dimensions)
        if len_cutout_dimensions > 0:
            result_hdu_list = None
            if self._require_primary_hdu(cutout_dimensions) and \
               hdu_list[0].header['NAXIS'] == 0:
                # add the PrimaryHDU from the original HDU list
                result_hdu_list = fits.HDUList([hdu_list[0]])
            # Check for a pixel cutout
            if isinstance(cutout_dimensions[0], PixelCutoutHDU):
                for cutout_dimension in cutout_dimensions:
                    ext = cutout_dimension.get_extension()
                    ext_idx = hdu_list.index_of(ext)
                    hdu = hdu_list[ext_idx]

                    # Entire extension was requested.
                    if not cutout_dimension.get_ranges():
                        logger.debug(
                            'Appending entire extension {}'.format(ext))
                        result_hdu_list = self._add_hdu(hdu, result_hdu_list)
                        has_match = True
                    else:
                        try:
                            result_hdu_list = self._add_hdu(
                                self._pixel_cutout(hdu, cutout_dimension),
                                result_hdu_list)
                            has_match = True
                            logger.debug(
                                'Successfully cutout from {} ({})'.format(
                                    ext, ext_idx))
                        except NoContentError:
                            logger.debug(
                                'Skipping non-overlapping cutout {}'.format(
                                    cutout_dimension))
            else:
                # Skip the primary as it should be written out already.
                for hdu in hdu_list:
                    if hdu.is_image and hdu.data is not None:
                        transform = Transform()
                        logger.debug(
                            'Transforming {}'.format(cutout_dimensions))
                        transformed_cutout_dimension = \
                            transform.world_to_pixels(
                                cutout_dimensions, hdu.header)
                        logger.debug('Transformed {} into {}'.format(
                            cutout_dimensions, transformed_cutout_dimension))
                        try:
                            result_hdu_list = self._add_hdu(self._pixel_cutout(
                                hdu,
                                transformed_cutout_dimension), result_hdu_list)
                            has_match = True
                        except NoContentError:
                            logger.debug(
                                'Skipping non-overlapping cutout {}'.format(
                                     cutout_dimensions))

            # time to write the cutout file
            result_hdu_list.writeto(
                self.output_writer, output_verify='ignore', checksum='remove')
        else:
            raise NoContentError('No overlap found (No cutout specified).')

        logger.debug('Has match in list? -- {}'.format(has_match))
        return has_match

    def _add_hdu(self, hdu, result_hdu_list):
        if result_hdu_list:
            result_hdu_list.append(hdu)
        else:
            primary_hdu = fits.PrimaryHDU(header=hdu.header, data=hdu.data)
            # astropy might remove BSCALE and BZERO. Put them back if true
            self._fix_header(primary_hdu.header, hdu.header)
            result_hdu_list = fits.HDUList([primary_hdu])
        return result_hdu_list

    def cutout(self, cutout_dimensions):
        # Start with the first extension
        hdu_list = fits.open(
            name=self.input_stream, memmap=True, mode='readonly',
            do_not_scale_image_data=True, ignore_missing_end=True)

        # Keep a tally of whether at least one HDU matched.
        has_match = self._check_hdu_list(cutout_dimensions, hdu_list)

        if not has_match:
            raise NoContentError('No overlap found.')



import logging

from cadccutout.core import *   # noqa
from cadccutout.cutoutnd import *   # noqa
from cadccutout.file_helper import *   # noqa
from cadccutout.no_content_error import *   # noqa
from cadccutout.pixel_cutout_hdu import *   # noqa
from cadccutout.pixel_range_input_parser import *   # noqa
from cadccutout.utils import *   # noqa
from cadccutout.file_helpers import *   # noqa
from cadccutout.shape import *   # noqa
from cadccutout.transform import *   # noqa

logging.basicConfig()
logger = logging.getLogger(__name__)



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
import re
from cadccutout.pixel_cutout_hdu import PixelCutoutHDU
from cadccutout.utils import to_num, is_integer

__all__ = ['PixelRangeInputParserError', 'PixelRangeInputParser']

logger = logging.getLogger(__name__)

RANGE_BEGIN_CHAR = '['
RANGE_END_CHAR = ']'


class PixelRangeInputParserError(ValueError):
    pass


class PixelRangeInputParser(object):
    """
    Parse the pixel input as it was delivered.
    """

    def __init__(self, delimiter=':', separator=','):
        self.delimiter = delimiter
        self.separator = separator
        self.match_pattern = re.compile(
            r'[\s*\[?[\w]*,?\s*\d*\]?]?[\[?[\d*:?\d*,?\s*]*\]?]')
        self.valid_range_pattern = r'\d*:?\d+'

    def is_pixel_cutout(self, input_str):
        return input_str and input_str.count(RANGE_BEGIN_CHAR) > 0

    def _is_valid_range(self, rs):
        match = re.search(self.valid_range_pattern, rs)
        return match and match.group()

    def _to_range_tuple(self, rs):
        if self.delimiter not in rs:
            if is_integer(rs):
                val = to_num(rs)
            else:
                val = rs
            return (val, val)  # Turns 7 into 7:7
        else:
            items = rs.split(self.delimiter)

            if not items:
                raise PixelRangeInputParserError(
                    'Incomplete range specified {}'.format(rs))
            else:
                sanitized_items = []
                for item in items:
                    if is_integer(item):
                        sanitized_items.append(to_num(item))
                    else:
                        sanitized_items.append(item)

                return tuple(sanitized_items)

    def parse(self, pixel_range_input_str):
        """
        Parse a string range.
        :param  pixel_range_input_str: The string to parse.
        :return List of PixelCutoutHDU instances

        Example:

        rp = PixelRangeInputParser()
        rp.parse('[0][1]')
        => [PixelCutoutHDU((1,1), extension='0')]

        rp.parse('[99:112]')
        => [PixelCutoutHDU((99,112), extension='0')]

        rp.parse('[SCI][99:112][5]')
        => [PixelCutoutHDU((99,112), extension=SCI),
            PixelCutoutHDU(extension='5')]

        rp.parse('[IMG,2][100:112][6][300:600]')
        => [PixelCutoutHDU((100,112), extension='IMG,2'),
            PixelCutoutHDU((300,600), extension='6')]
        """
        rs = pixel_range_input_str.strip()

        if not self.is_pixel_cutout(rs):
            raise PixelRangeInputParserError(
                'Not a valid pixel cutout string "{}".'.format(rs))

        # List of ranges in format [ext][pixel ranges]
        ranges = re.findall(self.match_pattern, rs)

        if not ranges:
            raise PixelRangeInputParserError(
                'Invalid range specified.  Should be in the format of {}  \
                (i.e.[0][8:35]), or single digit(i.e. 9), or single name (SCI).'
                .format(self.match_pattern))

        parsed_items = []

        for r in ranges:
            pixel_ranges = []
            logger.debug('Next range is {}'.format(r))
            split_items = list(map(lambda x: x.split(
                '[')[1], list(filter(None, r.split(']')))))
            l_items = len(split_items)

            if l_items == 2:
                extension = split_items[0]
                pixel_ranges = list(map(self._to_range_tuple, list(
                    filter(self._is_valid_range, split_items[1].split(
                        self.separator)))))
            elif l_items == 1:
                item = split_items[0].strip()
                if item.count(self.delimiter) > 0:
                    extension = '0'
                    pixel_ranges = list(map(self._to_range_tuple, list(
                        filter(self._is_valid_range,
                               item.split(self.separator)))))
                else:
                    extension = item
            else:
                raise PixelRangeInputParserError(
                    'Nothing usable for range {}'.format(r))

            parsed_items.append(PixelCutoutHDU(
                dimension_ranges=pixel_ranges, extension=extension))

        return parsed_items



# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2018.                            (c) 2018.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU AfferoF
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#  $Revision: 1 $
#
# ***********************************************************************
#

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging

from copy import deepcopy
from math import ceil

from astropy.wcs import Sip
from cadccutout.no_content_error import NoContentError

__all__ = ['CutoutResult', 'CutoutND']

logger = logging.getLogger(__name__)


class CutoutResult(object):
    """
    Just a DTO to move results of a cutout.  It's more readable than a plain
    tuple.
    """

    def __init__(self, data, wcs=None):
        self.data = data
        self.wcs = wcs


class CutoutND(object):
    def __init__(self, data, wcs=None):
        """
        Parameters
        ----------
        data : `~numpy.ndarray`
            The N-dimensional data array from which to extract the cutout array.
        cutout_region : `PixelCutoutHDU`
            The Pixel HDU Cutout description.  See
            cadccutout.pixel_cutout_hdu.py.
        wcs : `~astropy.wcs.WCS` or `None`
            A WCS object associated with the cutout array.  If it's specified,
            reset the WCS values for the cutout.

        Returns
        -------
        CutoutResult instance
        """
        if data is None:
            raise NoContentError('Nothing to cutout from.')

        self.data = data
        self.wcs = wcs

    def to_slice(self, idx, cutout_region):
        """
        Convert a region (tuple) into a slice to be used in a data array.  Some
        fenangling is added here to adjust the values based on a striding step,
        or to handle a two-dimensional cutout where y2 < y1.

        Some math is translated (or invented) to match what cfitsio (imcopy) is
        producing.

        Parameters
        ----------
        idx:        Index of the current region requested.  Used to lookup the
                    shape of the data to "pad" the cutout if necessary.
        cutout_region:  The requested region to be cutout.
        """
        len_region = len(cutout_region)

        if len_region > 0:
            low_bound = cutout_region[0]

            if low_bound == '*':
                lower_bound = 0
                upper_bound = self.data.shape[idx]
                step = int(cutout_region[1])
            else:
                lower_bound = int(low_bound)
                if lower_bound > 0:
                    lower_bound -= 1
                upper_bound = int(cutout_region[1])
                if (len_region == 3):
                    step = int(cutout_region[2])
                    if lower_bound > upper_bound:
                        upper_bound -= 2
                        if step > 0:
                            step *= -1
                elif lower_bound > upper_bound:
                    upper_bound -= 2
                    step = -1
                else:
                    step = 1

            logger.debug('Bounds are {}:{}:{}'.format(
                lower_bound, upper_bound, step))
            return slice(lower_bound, upper_bound, step)
        else:
            raise ValueError('Should have at least two values (lower, upper).')

    def _pad_cutout(self, cutout_shape):
        len_shape = len(cutout_shape)
        data_shape = self.data.shape
        logger.debug('Data shape is {} with length {}'.format(
            data_shape, len(self.data)))
        len_data = len(data_shape)
        if len_data > len_shape:
            missing_shape_bounds = data_shape[:len_data - len_shape]
            logger.debug('Missing shape bounds are {} for length {}'.format(
                missing_shape_bounds, len_data - len_shape))

            for val in missing_shape_bounds:
                cutout_shape.append(slice(val))

    def format_wcs(self, cutout_shape):
        """
        Re-calculate the CRPIX values for the WCS.  The SIP values are also
        re-calculated, if present.

        CRPIX values are tricky and there exists some black magic math in here,
        not to mention some values are set differently to accommodate the subtle
        variations with Python 2/3 float values.

        Parameters
        ----------
        cutout_shape:   The tuple containing the bounding values (shape) of the
        resulting cutout.

        Returns
        -------
        The formatted WCS object.
        """
        output_wcs = deepcopy(self.wcs)
        wcs_crpix = output_wcs.wcs.crpix
        l_wcs_crpix = len(wcs_crpix)

        logger.debug('Adjusting WCS.')

        for idx, cutout_region in enumerate(cutout_shape):
            if idx < l_wcs_crpix:
                curr_val = wcs_crpix[idx]
                start = cutout_region.start
                step = cutout_region.step

                if start:
                    wcs_crpix[idx] -= float(ceil(start + 0.5))

                if step is not None:
                    logger.debug('Taking step {} into account.'.format(step))
                    wcs_crpix[idx] /= step

                if start:
                    wcs_crpix[idx] += 1.0

                logger.debug(
                    'Adjusted wcs_crpix val from {} to {}'.format(
                        curr_val, wcs_crpix[idx]))

        if output_wcs.wcs.has_pc():
            pc = output_wcs.wcs.pc
            for i in range(output_wcs.wcs.naxis):
                for j in range(output_wcs.wcs.naxis):
                    step = cutout_shape[j].step
                    if step:
                        pc[i][j] *= step
        elif output_wcs.wcs.has_cd():
            cd = output_wcs.wcs.cd
            for i in range(output_wcs.wcs.naxis):
                for j in range(output_wcs.wcs.naxis):
                    step = cutout_shape[j].step
                    if step:
                        cd[i][j] *= step

        if self.wcs.sip:
            curr_sip = self.wcs.sip
            output_wcs.sip = Sip(curr_sip.a, curr_sip.b,
                                 curr_sip.ap, curr_sip.bp,
                                 wcs_crpix[0:2])

        logger.debug('WCS adjusted.')

        return output_wcs

    def extract(self, cutout_regions):
        """
        Perform the extraction from the data for the provided region.  If the
        provided region is smaller than the data, it will be padded with the
        values from the data.

        :param cutout_regions:    List of region tuples.
        """
        try:
            cutout_shape = [
                self.to_slice(idx, cutout_region) for idx, cutout_region in
                enumerate(cutout_regions)]
            self._pad_cutout(cutout_shape)

            cutout = tuple(reversed(cutout_shape))
            logger.debug('Cutout is {}'.format(cutout))
            cutout_data = self.data[cutout]
        except IndexError:
            raise NoContentError('No content (arrays do not overlap).')

        logger.debug('Extracted {} of data from {}.'.format(cutout_data.shape,
                                                            self.data.shape))

        if self.wcs:
            output_wcs = self.format_wcs(cutout_shape)
        else:
            logger.debug('No WCS present.')
            output_wcs = None

        return CutoutResult(data=cutout_data, wcs=output_wcs)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import glob
import os
import sys

from setuptools.command.test import test as TestCommand
from setuptools import find_packages

from setuptools import setup

import distutils.cmd

# read the README.rst file and return as string.


def readme():
    with open('README.rst') as r_obj:
        return r_obj.read()


# Get some values from the setup.cfg
try:
    from ConfigParser import ConfigParser
except ImportError:
    from configparser import ConfigParser

conf = ConfigParser()
conf.optionxform = str
conf.read(['setup.cfg'])
metadata = dict(conf.items('metadata'))

PACKAGENAME = metadata.get('package_name', 'packagename')
DESCRIPTION = metadata.get('description', 'CADC package')
AUTHOR = metadata.get('author', 'CADC')
AUTHOR_EMAIL = metadata.get('author_email', 'cadc@nrc.gc.ca')
LICENSE = metadata.get('license', 'unknown')
URL = metadata.get('url', 'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca')

# VERSION should be PEP386 compatible (http://www.python.org/dev/peps/pep-0386)
VERSION = metadata.get('version', 'none')

# generate the version file
with open(os.path.join(PACKAGENAME, 'version.py'), 'w') as f:
    f.write('version = \'{}\'\n'.format(VERSION))

# Treat everything in scripts except README.rst as a script to be installed
scripts = [fname for fname in glob.glob(os.path.join('scripts', '*'))
           if os.path.basename(fname) != 'README.rst']

# Define entry points for command-line scripts
entry_points = {'console_scripts': []}

entry_point_list = conf.items('entry_points')
for entry_point in entry_point_list:
    entry_points['console_scripts'].append('{0} = {1}'.format(entry_point[0],
                                                              entry_point[1]))

# add the --cov option to the test command


class PyTest(TestCommand):
    """
    Class py.test for the testing
    """
    user_options = []

    def __init__(self, dist, **kw):
        TestCommand.__init__(self, dist, **kw)
        self.pytest_args = ['--cov', PACKAGENAME, '--disable-warnings']

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        err_no = pytest.main(self.pytest_args)
        sys.exit(err_no)


class PyIntTest(TestCommand):
    """
    Class py.test for the int testing
    """
    user_options = []

    def __init__(self, dist, **kw):
        THIS_DIR = os.path.dirname(os.path.realpath(__file__))
        INT_TEST_DIR = os.path.join(THIS_DIR, PACKAGENAME, 'int_tests')
        TestCommand.__init__(self, dist, **kw)
        self.pytest_args = [str(INT_TEST_DIR)]

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        err_no = pytest.main(self.pytest_args)
        sys.exit(err_no)


install_requires = metadata.get('install_requires', '').strip().split()
all_install_requires = [install_requires]
dependency_links = metadata.get('dependency_links', '').strip().split()
runtime_env = os.getenv('ENV', 'PROD')

if runtime_env == 'DEV':
    all_install_requires.append(metadata.get(
        'dev_install_requires', '').strip().split())

setup(name=PACKAGENAME,
      version=VERSION,
      description=DESCRIPTION,
      scripts=scripts,
      install_requires=all_install_requires,
      dependency_links=dependency_links,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      url=URL,
      long_description=readme(),
      zip_safe=False,
      use_2to3=False,
      setup_requires=['pytest-runner'],
      entry_points=entry_points,
      packages=find_packages(),
      package_data={PACKAGENAME: ['data/*']},
      classifiers=[
          'Natural Language :: English',
          'License :: OSI Approved :: GNU Affero General Public License v3',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7'
      ],
      cmdclass={
          'coverage': PyTest, 'int_test': PyIntTest
      }
      )
